#===============================================================
# CodeC1-advChi.py
# A very simple Navier-Stokes solver for a drop falling in a
# rectangular box, using a conservative form of the equations. 
# A first-order explicit projection method and centered in space 
# discretizationa are used. The marker function is advected by 
# an advection diffusion equation.
# Original Matlab code by Gretar Tryggvason
# Python code converted by Tingyi Lu on 7/25/2018
#===============================================================

import numpy as np
import matplotlib.pyplot as plt
import time as time0

Lx=1.0;Ly=1.0;gx=0.0;gy=-100.0  # Domain size and
rho1=1.0; rho2=2.0; m0=0.01     # physical variables
unorth=0;usouth=0;veast=0;vwest=0;time=0.0 
rad=0.15;xc=0.5;yc=0.7 # Initial drop size and location

#-------------------- Numerical variables ----------------------
nx=32;ny=32;dt=0.00125;nstep=400; maxit=200; maxError=0.001; beta=1.5

#-------------------- Zero various arrys -----------------------
u=np.zeros((nx+2,ny+3));  v=np.zeros((nx+3,ny+2));  p=np.zeros((nx+3,ny+3))
ut=np.zeros((nx+2,ny+3)); vt=np.zeros((nx+3,ny+2)); tmp1=np.zeros((nx+3,ny+3)) 
uu=np.zeros((nx+2,ny+2)); vv=np.zeros((nx+2,ny+2)); tmp2=np.zeros((nx+3,ny+3))
r=np.zeros((nx+3,ny+3)); chi=np.zeros((nx+3,ny+3))

dx=Lx/nx;dy=Ly/ny                          # Set the grid 
x=np.linspace(-1.5*dx, (nx+0.5)*dx, nx+3)
y=np.linspace(-1.5*dy, (ny+0.5)*dy, ny+3)
xh=np.linspace(-dx, nx*dx, nx+2)
yh=np.linspace(-dy, ny*dy, ny+2)

X, Y = np.meshgrid(x,y)

#-------------------- Initial Conditions -----------------------
r = np.zeros((nx+3,nx+3))+rho1 # Set density

for i in range(nx+3):  
    for j in range(ny+3):                     # for the domain and the drop
        if((x[i]-xc)**2+(y[j]-yc)**2 <rad**2):
            r[i,j] = rho2
            chi[i,j] = 1.0
            
plt.contour(X,Y,chi.T)
plt.axis('scaled')
plt.axis([0, Lx, 0, Ly], aspect=1)

time_start = time0.time() 

#---------------------- START TIME LOOP ------------------------
for istep in range(nstep):
    print(istep)

    #--- ADVECT marker using centered difference plus diffusion ----
    chio=chi.copy()
    
    chi[2:nx+2,2:ny+2]=(chio[2:nx+2,2:ny+2]-(0.5*dt/dx)*(u[2:nx+2,2:ny+2]*(chio[3:nx+3,2:ny+2]
                    +chio[2:nx+2,2:ny+2]) - u[1:nx+1,2:ny+2]*(chio[1:nx+1,2:ny+2]+chio[2:nx+2,2:ny+2]))
                    -(0.5* dt/dy)*(v[2:nx+2,2:ny+2]*(chio[2:nx+2,3:ny+3]
                    +chio[2:nx+2,2:ny+2])-v[2:nx+2,1:ny+1]*(chio[2:nx+2,1:ny+1]+chio[2:nx+2,2:ny+2])  )
                    +(m0*dt/dx/dx)*(chio[3:nx+3,2:ny+2]-2.0*chio[2:nx+2,2:ny+2]+chio[1:nx+1,2:ny+2])
                    +(m0*dt/dy/dy)*(chio[2:nx+2,3:ny+3]-2.0*chio[2:nx+2,2:ny+2]+chio[2:nx+2,1:ny+1]))
    
    #-------------------- Update the density -----------------------
    ro=r.copy()
    r = rho1 + (rho2-rho1)*chi

    #------------- Set tangential velocity at boundaries -----------	     
    u[:,1]=2*usouth-u[:,2];u[:,ny+2]=2*unorth-u[:,ny+1]
    v[1,:]=2*vwest-v[2,:];v[nx+2,:]=2*veast-v[nx+1,:]

    ut[2:nx+1,2:ny+2]=((2.0/(r[3:nx+2,2:ny+2]+r[2:nx+1,2:ny+2]))*(0.5*
                           (ro[3:nx+2,2:ny+2]+ro[2:nx+1,2:ny+2])*u[2:nx+1,2:ny+2]+ dt* (
                -(0.25/dx)*(ro[3:nx+2,2:ny+2]*(u[3:nx+2,2:ny+2]+u[2:nx+1,2:ny+2])**2-
				            ro[2:nx+1,2:ny+2]*(u[2:nx+1,2:ny+2]+u[1:nx,2:ny+2])**2)
            -(0.0625/dy)*( (ro[2:nx+1,2:ny+2]+ro[3:nx+2,2:ny+2]+ro[2:nx+1,3:ny+3]+ro[3:nx+2,3:ny+3])*
                            (u[2:nx+1,3:ny+3]+u[2:nx+1,2:ny+2])*(v[3:nx+2,2:ny+2]+v[2:nx+1,2:ny+2]) 
                          -(ro[2:nx+1,2:ny+2]+ro[3:nx+2,2:ny+2]+ro[3:nx+2,1:ny+1]+ro[2:nx+1,1:ny+1])*(u[2:nx+1,2:ny+2]
                            +u[2:nx+1,1:ny+1])*(v[3:nx+2,1:ny+1]+v[2:nx+1,1:ny+1]))
                       +m0*((u[3:nx+2,2:ny+2]-2*u[2:nx+1,2:ny+2]+u[1:nx,2:ny+2])/dx**2+ 
                            (u[2:nx+1,3:ny+3]-2*u[2:nx+1,2:ny+2]+u[2:nx+1,1:ny+1])/dy**2)
                     + 0.5*(ro[3:nx+2,2:ny+2]+ro[2:nx+1,2:ny+2])*gx ) ))
           
    vt[2:nx+2,2:ny+1]=((2.0/(r[2:nx+2,3:ny+2]+r[2:nx+2,2:ny+1]))*(0.5*
                           (ro[2:nx+2,3:ny+2]+ro[2:nx+2,2:ny+1])*v[2:nx+2,2:ny+1]+ dt* (     
            -(0.0625/dx)*( (ro[2:nx+2,2:ny+1]+ro[3:nx+3,2:ny+1]+ro[3:nx+3,3:ny+2]+ro[2:nx+2,3:ny+2])*
                            (u[2:nx+2,2:ny+1]+u[2:nx+2,3:ny+2])*(v[2:nx+2,2:ny+1]+v[3:nx+3,2:ny+1]) 
                         - (ro[2:nx+2,2:ny+1]+ro[2:nx+2,3:ny+2]+ro[1:nx+1,3:ny+2]+ro[1:nx+1,2:ny+1])*
                            (u[1:nx+1,3:ny+2]+u[1:nx+1,2:ny+1])*(v[2:nx+2,2:ny+1]+v[1:nx+1,2:ny+1]) )                                 
                -(0.25/dy)*(ro[2:nx+2,3:ny+2]*(v[2:nx+2,3:ny+2]+v[2:nx+2,2:ny+1])**2-
                            ro[2:nx+2,2:ny+1]*(v[2:nx+2,2:ny+1]+v[2:nx+2,1:ny])**2 )
                       +m0*((v[3:nx+3,2:ny+1]-2*v[2:nx+2,2:ny+1]+v[1:nx+1,2:ny+1])/dx**2+
                            (v[2:nx+2,3:ny+2]-2*v[2:nx+2,2:ny+1]+v[2:nx+2,1:ny])/dy**2)
                     + 0.5*(ro[2:nx+2,3:ny+2]+ro[2:nx+2,2:ny+1])*gy  ) ))     

    #------------------ Solve the Pressure Equation ----------------    
    rt=r.copy(); lrg=1e20   # Compute source term and the coefficient for p[i,j]
    rt[:,1]=lrg;rt[:,ny+2]=lrg
    rt[1,:]=lrg;rt[nx+2,:]=lrg


    tmp1[2:nx+2,2:ny+2]= ((0.5/dt)*( (ut[2:nx+2,2:ny+2]-ut[1:nx+1,2:ny+2])/dx+
                                     (vt[2:nx+2,2:ny+2]-vt[2:nx+2,1:ny+1])/dy ))
    tmp2[2:nx+2,2:ny+2]=(1.0/( (1./dx)*(1./(dx*(rt[3:nx+3,2:ny+2]+rt[2:nx+2,2:ny+2]))+   
                                        1./(dx*(rt[1:nx+1,2:ny+2]+rt[2:nx+2,2:ny+2])) )+ 
                               (1./dy)*(1./(dy*(rt[2:nx+2,3:ny+3]+rt[2:nx+2,2:ny+2]))+   
                                        1./(dy*(rt[2:nx+2,1:ny+1]+rt[2:nx+2,2:ny+2])) )   ))  

    for it in range(maxit):                # Solve for pressure by SOR
        oldp=p.copy()
    
        #Red & Black SOR
        for ipass in range(2):
            rb = ipass
            p[2+rb:nx+2:2,2:ny+2:2] = ( (1.0-beta)*p[2+rb:nx+2:2,2:ny+2:2] + beta*tmp2[2+rb:nx+2:2,2:ny+2:2]*(
                         (1.0/dx/dx)* ( p[3+rb:nx+3:2,2:ny+2:2]/(rt[3+rb:nx+3:2,2:ny+2:2]+rt[2+rb:nx+2:2,2:ny+2:2])
                                       +p[1+rb:nx+1:2,2:ny+2:2]/(rt[1+rb:nx+1:2,2:ny+2:2]+rt[2+rb:nx+2:2,2:ny+2:2])) 
                        +(1.0/dy/dy)* ( p[2+rb:nx+2:2,3:ny+3:2]/(rt[2+rb:nx+2:2,3:ny+3:2]+rt[2+rb:nx+2:2,2:ny+2:2])
                                       +p[2+rb:nx+2:2,1:ny+1:2]/(rt[2+rb:nx+2:2,1:ny+1:2]+rt[2+rb:nx+2:2,2:ny+2:2]))
                        - tmp1[2+rb:nx+2:2,2:ny+2:2] ) )

            rb=1-ipass
            p[2+rb:nx+2:2,3:ny+2:2] = ( (1.0-beta)*p[2+rb:nx+2:2,3:ny+2:2] + beta*tmp2[2+rb:nx+2:2,3:ny+2:2]*(
                         (1.0/dx/dx)* ( p[3+rb:nx+3:2,3:ny+2:2]/(rt[3+rb:nx+3:2,3:ny+2:2]+rt[2+rb:nx+2:2,3:ny+2:2])
                                       +p[1+rb:nx+1:2,3:ny+2:2]/(rt[1+rb:nx+1:2,3:ny+2:2]+rt[2+rb:nx+2:2,3:ny+2:2])) 
                        +(1.0/dy/dy)* ( p[2+rb:nx+2:2,4:ny+3:2]/(rt[2+rb:nx+2:2,4:ny+3:2]+rt[2+rb:nx+2:2,3:ny+2:2])
                                       +p[2+rb:nx+2:2,2:ny+1:2]/(rt[2+rb:nx+2:2,2:ny+1:2]+rt[2+rb:nx+2:2,3:ny+2:2]))
                        - tmp1[2+rb:nx+2:2,3:ny+2:2] ) )
         
        p[1,:] = p[2,:]; p[nx+2,:] = p[nx+1,:]
        p[:,1] = p[:,2]; p[:,ny+2] = p[:,ny+1]

        if (np.abs(oldp-p)).max() < maxError : 
            break

    # Correct the u-velocity 
    u[2:nx+1,2:ny+2]=(ut[2:nx+1,2:ny+2]-dt*(2.0/dx)*(p[3:nx+2,2:ny+2]-p[2:nx+1,2:ny+2])/(r[3:nx+2,2:ny+2]+r[2:nx+1,2:ny+2]))
    
    # Correct the v-velocity
    v[2:nx+2,2:ny+1]=(vt[2:nx+2,2:ny+1]-dt*(2.0/dy)*(p[2:nx+2,3:ny+2]-p[2:nx+2,2:ny+1])/(r[2:nx+2,3:ny+2]+r[2:nx+2,2:ny+1]))
    
    #------------------ Plot the results ---------------------------
    time=time+dt                   # plot the results
    uu[1:nx+2,1:ny+2]=(0.5*(u[1:nx+2,2:ny+3]+u[1:nx+2,1:ny+2]))
    vv[1:nx+2,1:ny+2]=(0.5*(v[2:nx+3,1:ny+2]+v[1:nx+2,1:ny+2]))

    plt.cla()
    plt.contour(x[2:nx+2],y[2:ny+2],chi.T[2:nx+2,2:ny+2])
    plt.quiver(xh[1:],yh[1:],uu.T[1:,1:],vv.T[1:,1:])

    plt.pause(0.01)
	
print('time_elapsed= %s' % (time0.time() - time_start) )

plt.close()

                  # End of time step