#===============================================================
# CodeC3-frt-st-RK3.m
# A very simple Navier-Stokes solver for a drop falling in a
# rectangular box, using a conservative form of the equations. 
# A 3-order explicit projection method and centered in space 
# discretizationa are used. The density is advected by a front 
# tracking scheme and surface tension and variable viscosity is
# included. This version uses a simple method to create the 
# marker function.
# Original Matlab code by Gretar Tryggvason
# Python code converted by Tingyi Lu on 7/25/2018
#===============================================================

import numpy as np
import matplotlib.pyplot as plt
import time as time0

Lx=1.0;Ly=1.0;gx=0.0;gy=-100.0; sigma=10 # Domain size and
rho1=1.0; rho2=2.0; m1=0.01; m2=0.02     # physical variables
unorth=0; usouth=0; veast=0; vwest=0; time=0.0 
rad=0.15; xc=0.5; yc=0.7          # Initial drop size and location

#-------------------- Numerical variables ----------------------
nx=32;ny=32;dt=0.001;nstep=400; maxit=200;maxError=0.001;beta=1.5; Nf=100; Maxf=2000

#-------------------- Zero various arrys -----------------------
u=np.zeros((nx+1,ny+2));  v=np.zeros((nx+2,ny+1));  p=np.zeros((nx+2,ny+2))
ut=np.zeros((nx+1,ny+2)); vt=np.zeros((nx+2,ny+1)); tmp1=np.zeros((nx+2,ny+2)) 
uu=np.zeros((nx+1,ny+1)); vv=np.zeros((nx+1,ny+1)); tmp2=np.zeros((nx+2,ny+2))
fx=np.zeros((nx+2,ny+2)); fy=np.zeros((nx+2,ny+2)); r=np.zeros((nx+2,ny+2))
r=np.zeros((nx+2,ny+2));  chi=np.zeros((nx+2,ny+2)) 
m=np.zeros((nx+2,ny+2));  d=np.zeros((nx+2,ny+2))
xf=np.zeros(Maxf); yf=np.zeros(Maxf) 
uf=np.zeros(Maxf); vf=np.zeros(Maxf)
tx=np.zeros(Maxf); ty=np.zeros(Maxf)
un=np.zeros((nx+1,ny+2)); vn=np.zeros((nx+2,ny+1))   # Used for 
rn=np.zeros((nx+2,ny+2)); mn=np.zeros((nx+2,ny+2))   # higher order
xfn=np.zeros(Maxf); yfn=np.zeros(Maxf)               # in time
Area=np.zeros(nstep);CentroidX=np.zeros(nstep);CentroidY=np.zeros(nstep)
Time1=np.zeros(nstep)


dx=Lx/nx;dy=Ly/ny                                    # Set the grid 
x=np.linspace(-.5*dx, (nx+0.5)*dx, nx+2)
y=np.linspace(-.5*dx, (ny+0.5)*dy, ny+2)
xh=np.linspace(0, Lx, nx+1)
yh=np.linspace(0, Ly, ny+1)

#-------------------- Initial Conditions -----------------------
r[:,:]=rho1;m[:,:]=m1 # Set density and viscosity

for i in range(1,nx+1):                     # for the domain and the drop
    for j in range(1,ny+1):
        if((x[i]-xc)**2+(y[j]-yc)**2 < rad**2):
            r[i,j]=rho2
            m[i,j]=m2
            chi[i,j]=1.0

for l in range(Nf+2):
    xf[l]=xc-rad*np.sin(2.0*np.pi*l/Nf)      # Initialize 
    yf[l]=yc+rad*np.cos(2.0*np.pi*l/Nf)      # the Front

plt.plot(xf[0:Nf],yf[0:Nf],'k',linewidth=3)
plt.axis('scaled')
plt.axis([0,Lx,0,Ly], aspect=1)
plt.pause(0.0001)

time_start = time0.time()                                          

#---------------------- START TIME LOOP ------------------------
for istep in range(nstep):
    print(istep)
    un=u.copy(); vn=v.copy(); rn=r.copy(); mn=m.copy(); xfn=xf.copy(); yfn=yf.copy()  # Higher order
    for substep in range(3):                          # in time

#---------------------- Advect the Front -----------------------
        for l in range(1,Nf+1):                       # Interpolate the Front Velocities
            ip=np.int(xf[l]/dx); jp=np.int((yf[l]+0.5*dy)/dy)
            ax=xf[l]/dx-ip; ay=(yf[l]+0.5*dy)/dy-jp	   
            uf[l]=((1.0-ax)*(1.0-ay)*u[ip,jp]+ax*(1.0-ay)*u[ip+1,jp]+
		             (1.0-ax)*ay*u[ip,jp+1]+ax*ay*u[ip+1,jp+1])
						
            ip=np.int((xf[l]+0.5*dx)/dx); jp=np.int(yf[l]/dy)
            ax=(xf[l]+0.5*dx)/dx-ip; ay=yf[l]/dy-jp
            vf[l]=((1.0-ax)*(1.0-ay)*v[ip,jp]+ax*(1.0-ay)*v[ip+1,jp]+
		             (1.0-ax)*ay*v[ip,jp+1]+ax*ay*v[ip+1,jp+1])		
		
        xf[1:Nf+1]+=dt*uf[1:Nf+1]
        yf[1:Nf+1]+=dt*vf[1:Nf+1] # Move the
        xf[0]=xf[Nf];yf[0]=yf[Nf];xf[Nf+1]=xf[1];yf[Nf+1]=yf[1] # Front        
    
 	
#-------------- Update the marker function ---------------------
        d[:,:]=2

        for l in range(1,Nf+1):
            nfx=-(yf[l+1]-yf[l])/dx   
            nfy=(xf[l+1]-xf[l])/dy  # Normal vector
            ds=np.sqrt(nfx*nfx+nfy*nfy); nfx=nfx/ds; nfy=nfy/ds
            xfront=0.5*(xf[l]+xf[l+1]); yfront=0.5*(yf[l]+yf[l+1])
            ip=np.int((xfront+0.5*dx)/dx); jp=np.int((yfront+0.5*dy)/dy)
    
            d1=np.sqrt(((xfront-  x[ip])/dx)**2+((yfront-  y[jp])/dy)**2)
            d2=np.sqrt(((xfront-x[ip+1])/dx)**2+((yfront-  y[jp])/dy)**2)
            d3=np.sqrt(((xfront-x[ip+1])/dx)**2+((yfront-y[jp+1])/dy)**2)
            d4=np.sqrt(((xfront-  x[ip])/dx)**2+((yfront-y[jp+1])/dy)**2)
    
            if d1<d[ip,jp]:
                d[ip,jp]=d1.copy()
                dn1=(x[ip]-xfront)*nfx/dx+(y[jp]-yfront)*nfy/dy
                chi[ip,jp]=0.5*(1.0+np.sign(dn1)) 
                if abs(dn1)<0.5:
                    chi[ip,jp]=0.5+dn1
    
            if d2<d[ip+1,jp]:
                d[ip+1,jp]=d2.copy()
                dn2=(x[ip+1]-xfront)*nfx/dx+(y[jp]-yfront)*nfy/dy
                chi[ip+1,jp]=0.5*(1.0+np.sign(dn2))     
                if abs(dn2)<0.5:
                    chi[ip+1,jp]=0.5+dn2
            
            if d3<d[ip+1,jp+1]:  
                d[ip+1,jp+1]=d3.copy()
                dn3=(x[ip+1]-xfront)*nfx/dx+(y[jp+1]-yfront)*nfy/dy
                chi[ip+1,jp+1]=0.5*(1.0+np.sign(dn3)) 
                if abs(dn3)<0.5:
                    chi[ip+1,jp+1]=0.5+dn3
    
            if d4<d[ip,jp+1]:
                d[ip,jp+1]=d4.copy()
                dn4=(x[ip]-  xfront)*nfx/dx+(y[jp+1]-yfront)*nfy/dy
                chi[ip,jp+1]=  0.5*(1.0+np.sign(dn4))
                if abs(dn4)<0.5:
                    chi[ip,jp+1]=  0.5+dn4                
       
#-------------------- Update the density -----------------------
        ro=r.copy()
        r = rho1+(rho2-rho1)*chi

#------------------ Find surface tension -----------------------
        fx=np.zeros((nx+2,ny+2));fy=np.zeros((nx+2,ny+2))  # Set fx & fy to zero

        for l in range(Nf+1): 
            ds=np.sqrt((xf[l+1]-xf[l])**2+(yf[l+1]-yf[l])**2)
            tx[l]=(xf[l+1]-xf[l])/ds
            ty[l]=(yf[l+1]-yf[l])/ds # Tangent vectors
    
        tx[Nf+1]=tx[1];ty[Nf+1]=ty[1]
	
        for l in range(1,Nf+1):           # Distribute to the fixed grid
            nfx=sigma*(tx[l]-tx[l-1]);nfy=sigma*(ty[l]-ty[l-1])
        
            ip=np.int(xf[l]/dx); jp=np.int((yf[l]+0.5*dy)/dy)
            ax=xf[l]/dx-ip; ay=(yf[l]+0.5*dy)/dy-jp
            fx[ip,jp]    +=(1.0-ax)*(1.0-ay)*nfx/dx/dy
            fx[ip+1,jp]  +=ax*(1.0-ay)*nfx/dx/dy
            fx[ip,jp+1]  +=(1.0-ax)*ay*nfx/dx/dy
            fx[ip+1,jp+1]+=ax*ay*nfx/dx/dy
      
            ip=np.int((xf[l]+0.5*dx)/dx); jp=np.int(yf[l]/dy)
            ax=(xf[l]+0.5*dx)/dx-ip; ay=yf[l]/dy-jp	  
            fy[ip,jp]    +=(1.0-ax)*(1.0-ay)*nfy/dx/dy
            fy[ip+1,jp]  +=ax*(1.0-ay)*nfy/dx/dy
            fy[ip,jp+1]  +=(1.0-ax)*ay*nfy/dx/dy
            fy[ip+1,jp+1]+=ax*ay*nfy/dx/dy  		

        fx[:,1]=fx[:,1]+fx[:,0]           # Correct boundary
        fx[:,ny]=fx[:,ny]+fx[:,ny+1]  # values for the
        fy[1,:]=fy[1,:]+fy[0,:]           # surface force
        fy[nx,:]=fy[nx,:]+fy[nx+1,:]  # on the grid
#------------- Set tangential velocity at boundaries -----------	     
        u[:,0]=2*usouth-u[:,1];u[:,ny+1]=2*unorth-u[:,ny]
        v[0,:]=2*vwest-v[1,:];v[nx+1,:]=2*veast-v[nx,:]


#-------------- Find the predicted velocities ------------------	     

                                # Temporary u-velocity-advection
        ut[1:nx,1:ny+1]=( (2.0/(r[2:nx+1,1:ny+1]+r[1:nx,1:ny+1]))
		                 *(0.5*(ro[2:nx+1,1:ny+1]+ro[1:nx,1:ny+1])*u[1:nx,1:ny+1]+ dt* (
    			  -(0.25/dx)*(  ro[2:nx+1,1:ny+1]*(u[2:nx+1,1:ny+1]+u[1:nx,1:ny+1])**2
				               -ro[1:nx,1:ny+1]*(u[1:nx,1:ny+1]+u[:nx-1,1:ny+1])**2)
    			-(0.0625/dy)*( (ro[1:nx,1:ny+1]+ro[2:nx+1,1:ny+1]+ro[1:nx,2:ny+2]+ro[2:nx+1,2:ny+2])*
    			                (u[1:nx,2:ny+2]+u[1:nx,1:ny+1])*(v[2:nx+1,1:ny+1]+v[1:nx,1:ny+1]) 
    			              -(ro[1:nx,1:ny+1]+ro[2:nx+1,1:ny+1]+ro[2:nx+1,:ny]+ro[1:nx,:ny])*
							    (u[1:nx,1:ny+1]+u[1:nx,:ny])*(v[2:nx+1,:ny]+v[1:nx,:ny]))
    			         + 0.5*(ro[2:nx+1,1:ny+1]+ro[1:nx,1:ny+1])*gx + fx[1:nx,1:ny+1]) ))

							   # Temporary v-velocity-advection 
        vt[1:nx+1,1:ny]=((2.0/(r[1:nx+1,2:ny+1]+r[1:nx+1,1:ny]))*
                        (0.5*(ro[1:nx+1,2:ny+1]+ro[1:nx+1,1:ny])*v[1:nx+1,1:ny]+ dt* (     
              -(0.0625/dx)*( (ro[1:nx+1,1:ny]+ro[2:nx+2,1:ny]+ro[2:nx+2,2:ny+1]+ro[1:nx+1,2:ny+1])*
                              (u[1:nx+1,1:ny]+u[1:nx+1,2:ny+1])*(v[1:nx+1,1:ny]+v[2:nx+2,1:ny]) 
                           - (ro[1:nx+1,1:ny]+ro[1:nx+1,2:ny+1]+ro[:nx,2:ny+1]+ro[:nx,1:ny])*
                              (u[:nx,2:ny+1]+u[:nx,1:ny])*(v[1:nx+1,1:ny]+v[:nx,1:ny]) )                                 
                  -(0.25/dy)*(ro[1:nx+1,2:ny+1]*(v[1:nx+1,2:ny+1]+v[1:nx+1,1:ny])**2-
                              ro[1:nx+1,1:ny]*(v[1:nx+1,1:ny]+v[1:nx+1,:ny-1])**2 )
                       + 0.5*(ro[1:nx+1,2:ny+1]+ro[1:nx+1,1:ny])*gy + fy[1:nx+1,1:ny] ) ))     

					  
					#Temporary u-velocity-viscosity
        ut[1:nx,1:ny+1]=(ut[1:nx,1:ny+1]+(2.0/(r[2:nx+1,1:ny+1]+r[1:nx,1:ny+1]))*dt*(                                         
             +(1./dx)*2.*(m[2:nx+1,1:ny+1]*(1./dx)*(u[2:nx+1,1:ny+1]-u[1:nx,1:ny+1]) -       
                          m[1:nx,1:ny+1]  *(1./dx)*(u[1:nx,1:ny+1]-u[0:nx-1,1:ny+1]) )                 
         +(1./dy)*( 0.25*(m[1:nx,1:ny+1]+m[2:nx+1,1:ny+1]+m[2:nx+1,2:ny+2]+m[1:nx,2:ny+2])*        
                ((1./dy)*(u[1:nx,2:ny+2]-u[1:nx,1:ny+1]) + (1./dx)*(v[2:nx+1,1:ny+1]-v[1:nx,1:ny+1]) ) -  
                    0.25*(m[1:nx,1:ny+1]+m[2:nx+1,1:ny+1]+m[2:nx+1,0:ny]+m[1:nx,0:ny])*            
                ((1./dy)*(u[1:nx,1:ny+1]-u[1:nx,0:ny])+ (1./dx)*(v[2:nx+1,0:ny]- v[1:nx,0:ny])) ) )) 

					#Temporary v-velocity-viscosity
        vt[1:nx+1,1:ny]=(vt[1:nx+1,1:ny]+(2.0/(r[1:nx+1,2:ny+1]+r[1:nx+1,1:ny]))*dt*(
               +(1./dx)*( 0.25*(m[1:nx+1,1:ny]+m[2:nx+2,1:ny]+m[2:nx+2,2:ny+1]+m[1:nx+1,2:ny+1])*     
               ((1./dy)*(u[1:nx+1,2:ny+1]-u[1:nx+1,1:ny]) + (1./dx)*(v[2:nx+2,1:ny]-v[1:nx+1,1:ny]) ) -  
                          0.25*(m[1:nx+1,1:ny]+m[1:nx+1,2:ny+1]+m[0:nx,2:ny+1]+m[0:nx,1:ny])*            
               ((1./dy)*(u[0:nx,2:ny+1]-u[0:nx,1:ny])+ (1./dx)*(v[1:nx+1,1:ny]- v[0:nx,1:ny])) )
               +(1./dy)*2.*(m[1:nx+1,2:ny+1]*(1./dy)*(v[1:nx+1,2:ny+1]-v[1:nx+1,1:ny]) -           
                            m[1:nx+1,1:ny] *(1./dy)*(v[1:nx+1,1:ny]-v[1:nx+1,0:ny-1]) ) ))     

#------------------ Solve the Pressure Equation ----------------    
        rt=r.copy(); lrg=1000   # Compute source term and the coefficient for p(i,j)
        rt[:,0]=lrg;rt[:,ny+1]=lrg
        rt[0,:]=lrg;rt[nx+1,:]=lrg

        tmp1[1:nx+1,1:ny+1]=((0.5/dt)*( (ut[1:nx+1,1:ny+1]-ut[0:nx,1:ny+1])/dx+
                                        (vt[1:nx+1,1:ny+1]-vt[1:nx+1,0:ny])/dy))
        tmp2[1:nx+1,1:ny+1]=(1.0/( (1./dx)*(1./(dx*(rt[2:nx+2,1:ny+1]+rt[1:nx+1,1:ny+1]))+  
                                    1./(dx*(rt[0:nx,1:ny+1]+rt[1:nx+1,1:ny+1])) )+ 
                           (1./dy)*(1./(dy*(rt[1:nx+1,2:ny+2]+rt[1:nx+1,1:ny+1]))+   
                                    1./(dy*(rt[1:nx+1,0:ny]+rt[1:nx+1,1:ny+1])) )   ))
								   

        for it in range(maxit):	               # Solve for pressure by SOR
            oldArray=p.copy()

		        #Red & Black SOR
            for ipass in range(2):
                rb = ipass
                p[1+rb:nx+1:2,1:ny+1:2] = ( (1.0-beta)*p[1+rb:nx+1:2,1:ny+1:2] + beta*tmp2[1+rb:nx+1:2,1:ny+1:2]*(
                             (1.0/dx/dx)* ( p[2+rb:nx+2:2,1:ny+1:2]/(rt[2+rb:nx+2:2,1:ny+1:2]+rt[1+rb:nx+1:2,1:ny+1:2])
                                           +p[rb:nx:2,1:ny+1:2]/(rt[rb:nx:2,1:ny+1:2]+rt[1+rb:nx+1:2,1:ny+1:2])) 
                            +(1.0/dy/dy)* ( p[1+rb:nx+1:2,2:ny+2:2]/(rt[1+rb:nx+1:2,2:ny+2:2]+rt[1+rb:nx+1:2,1:ny+1:2])
                                           +p[1+rb:nx+1:2,0:ny:2]/(rt[1+rb:nx+1:2,0:ny:2]+rt[1+rb:nx+1:2,1:ny+1:2]))
                            - tmp1[1+rb:nx+1:2,1:ny+1:2] ) )
    
                rb=1-ipass
                p[1+rb:nx+1:2,2:ny+1:2] = ( (1.0-beta)*p[1+rb:nx+1:2,2:ny+1:2] + beta*tmp2[1+rb:nx+1:2,2:ny+1:2]*(
                             (1.0/dx/dx)* ( p[2+rb:nx+2:2,2:ny+1:2]/(rt[2+rb:nx+2:2,2:ny+1:2]+rt[1+rb:nx+1:2,2:ny+1:2])
                                           +p[rb:nx:2,2:ny+1:2]/(rt[rb:nx:2,2:ny+1:2]+rt[1+rb:nx+1:2,2:ny+1:2])) 
                            +(1.0/dy/dy)* ( p[1+rb:nx+1:2,3:ny+2:2]/(rt[1+rb:nx+1:2,3:ny+2:2]+rt[1+rb:nx+1:2,2:ny+1:2])
                                           +p[1+rb:nx+1:2,1:ny:2]/(rt[1+rb:nx+1:2,1:ny:2]+rt[1+rb:nx+1:2,2:ny+1:2]))
                            - tmp1[1+rb:nx+1:2,2:ny+1:2] ) )
    
            p[0,:] = p[1,:]; p[nx+1,:] = p[nx,:]
            p[:,0] = p[:,1]; p[:,ny+1] = p[:,ny]


            if (np.abs(oldArray-p)).max() <maxError:
                break

                # Correct the u-velocity 
        u[1:nx,1:ny+1]=ut[1:nx,1:ny+1]-dt*(2.0/dx)*(p[2:nx+1,1:ny+1]-p[1:nx,1:ny+1])/(r[2:nx+1,1:ny+1]+r[1:nx,1:ny+1])
               # Correct the v-velocity
        v[1:nx+1,1:ny]=vt[1:nx+1,1:ny]-dt*(2.0/dy)*(p[1:nx+1,2:ny+1]-p[1:nx+1,1:ny])/(r[1:nx+1,2:ny+1]+r[1:nx+1,1:ny])
		        #Update the viscosity
        m[0:nx+1,0:ny+2]=m1+(m2-m1)*chi[0:nx+1,0:ny+2]
				

        if substep==1: # Higher order (RK-3) in time
            u=0.75*un+0.25*u; v=0.75*vn+0.25*v; r=0.75*rn+0.25*r
            m=0.75*mn+0.25*m; xf=0.75*xfn+0.25*xf; yf=0.75*yfn+0.25*yf
        elif substep==2:
            u=(1/3)*un+(2/3)*u; v=(1/3)*vn+(2/3)*v; r=(1/3)*rn+(2/3)*r;
            m=(1/3)*mn+(2/3)*m; xf=(1/3)*xfn+(2/3)*xf; yf=(1/3)*yfn+(2/3)*yf;

#--------------- Add and deleate points in the Front -----------
    xfold=xf.copy();yfold=yf.copy(); j=0
    for l in range(1,Nf+1):
        ds=np.sqrt( ((xfold[l]-xf[j])/dx)**2 + ((yfold[l]-yf[j])/dy)**2)
        if ds > 0.5:
            j=j+1;xf[j]=0.5*(xfold[l]+xf[j-1]);yf[j]=0.5*(yfold[l]+yf[j-1])
            j=j+1;xf[j]=xfold[l];yf[j]=yfold[l]
        elif 0.25<=ds<=0.5:
            j=j+1;xf[j]=xfold[l];yf[j]=yfold[l]

    Nf=j-1
    xf[0]=xf[Nf];yf[0]=yf[Nf];xf[Nf+1]=xf[1];yf[Nf+1]=yf[1]
  
#----------------- Compute Diagnostic quantitites --------------
    Area[istep]=0; CentroidX[istep]=0; CentroidY[istep]=0; Time1[istep]=time 

    for l in range(1,Nf+1):
        Area[istep]+=0.25*((xf[l+1]+xf[l])*(yf[l+1]-yf[l])-(yf[l+1]+yf[l])*(xf[l+1]-xf[l]))  
        CentroidX[istep]+=0.125*((xf[l+1]+xf[l])**2+(yf[l+1]+yf[l])**2)*(yf[l+1]-yf[l])
        CentroidY[istep]-=0.125*((xf[l+1]+xf[l])**2+(yf[l+1]+yf[l])**2)*(xf[l+1]-xf[l])

    CentroidX[istep]=CentroidX[istep]/Area[istep];CentroidY[istep]=CentroidY[istep]/Area[istep]

#------------------ Plot the results ---------------------------
    time+=dt                   # plot the results
    uu[0:nx+1,0:ny+1]=0.5*(u[0:nx+1,1:ny+2]+u[0:nx+1,0:ny+1])
    vv[0:nx+1,0:ny+1]=0.5*(v[1:nx+2,0:ny+1]+v[0:nx+1,0:ny+1])
    
    plt.cla()
    plt.contour(x[1:nx+1],y[1:ny+1],chi.T[1:nx+1,1:ny+1])
	
    plt.quiver(xh,yh,uu.T,vv.T)

    plt.plot(xf[0:Nf],yf[0:Nf],'k',linewidth=3)

    plt.axis([0,Lx,0,Ly], aspect=1)
    plt.pause(0.0001)

print('time_elapsed= %s' % (time0.time() - time_start) )

#------ Extra commands for interactive processing --------------
#plt.figure(2)
#plt.plot(Time1,Area,'r',linewidth=2)
#plt.axis([0,dt*nstep,0,0.1])
#plt.show()
